/*
 * ListEdit.c - LineList manipulations
 *
 * The LineList structure is set up to allow an arbitrary number of
 * fields in each line.  Each field contains some text, in string
 * form.  This means that whatever kind of structure underlies the
 * line must be mapped onto the strings externally to the routines
 * in this file.
 *
 * All routines that operate on LineList objects assume
 * that there's at least one field in each line.
 */

# include	"ListEdit.h"


# define	normalHilite	0
# define	dimHilite		255


static LListPtr	trackList;
static short	trackPart;


/* ---------------------------------------------------------------- */
/*		Line list initialization, line creation and disposal		*/
/* ---------------------------------------------------------------- */


/*
 * Initialize a line list.  It's assumed that everything that
 * can be statically initialized already has been.  What's done
 * here:
 *
 * o	set the port
 * o	create the scroll bar
 * o	get a line list big enough for the given number of lines
 *
 * The port must be set before InitList is called, since the control
 * calculations depend on that.
 */

void
InitList (LListPtr theList, short maxLines)
{
GrafPtr	curPort;
Rect	r;
short	i;

	theList->lines = (LAHandle) NewHandle (maxLines * sizeof (LineHandle));
	theList->maxLines = maxLines;
	theList->hilite = false;
	r = theList->textRect;
	r.left = r.right + 1;
	r.right = r.left + 14;
	InsetRect (&r, -1, -1);
	GetPort (&curPort);
	theList->port = curPort;
	theList->scroll = NewControl (curPort, &r, "\p", true, 0, 0, 0,
									scrollBarProc, 0L);
}


/*
 * Create a line for a LineList object.  Must pass the number of
 * fields in a line.  NewLine doesn't know anything about the owner
 * of the line except that.  In particular, it doesn't attach the
 * line to the LineList.
 */

LineHandle
NewLine (short nFields)
{
FieldHandle	hField1,
			hField2 = nil;

	while (nFields-- > 0)
	{
		hField1 = New (Field);
		(**hField1).fStr[0] = 0;
		(**hField1).fNext = hField2;
		hField2 = hField1;
	}
	return (hField1);
}


/*
 * Dispose of a LineList line.  DisposeLine doesn't know anything
 * about the owner of the line.  In particular, it doesn't detach
 * the line from the LineList.
 *
 * The case of theLine = nil is handled w/o special treatment.
 */

void
DisposeLine (LListPtr theList, short n)
{
FieldHandle	hField1,
			hField2;

	hField1 = ListLine (theList, n);
	ListLine (theList, n) = nil;
	while (hField1 != nil)
	{
		hField2 = (**hField1).fNext;
		DisposeHandle ((Handle) hField1);
		hField1 = hField2;
	}
}


/*
 * Set the text of a field
 */

void
SetFieldStr (FieldHandle field, StringPtr str)
{
	HLock ((Handle) field);
	BlockMove (str, (**field).fStr, (long) (str[0] + 1));
	HUnlock ((Handle) field);
}


/*
 * Return number of the line in which point is located.  First
 * line is line zero.  This does *not* (and *shouldn't*) check
 * whether the point is actually in the text rectangle.
 */

short
FindLine (LListPtr theList, Point pt)
{
short	i, height;

	height = theList->lineHeight;
	i = pt.v - theList->textRect.top;
	if (i < 0)
		i -= height - 1;		/* so next division works properly */
	i /= height;
	return (i + theList->topVisLine);
}


/*
 * Scroll a LineList object absolutely to line n, i.e., so line
 * n is the top visible line.
 * This does not set the current line, since the lists are only
 * scrolled - nothing is actually selected.
 */

void
ScrollToLine (LListPtr theList, short n)
{
short	nLines;
short	topLine;
short	bottomLine;
short	visLines;
short	scrollAmt;
short	firstRedraw, redrawCount;
short	scrollFrom;

	nLines = theList->nLines;
	if (nLines == 0)
		return;				/* nothing to scroll to */

	topLine = theList->topVisLine;
	visLines = theList->visLines;
	bottomLine = topLine + visLines - 1;

	/*
	 * clip n to range 0..nLines-1
	 */
	if (n < 0)
		n = 0;
	if (n >= nLines)
		n = nLines - 1;
	
	/*
	 * Determine scroll direction
	 */
	if (n < topLine)
		scrollAmt = topLine - n;		/* toward beginning */
	else if (n > bottomLine)
		scrollAmt = bottomLine - n;	/* toward end */
	else
		scrollAmt = 0;				/* already visible, no scroll */

	if (nLines >= visLines)		/* enough to fill window */
	{
		if (nLines - visLines < topLine - scrollAmt)
			scrollAmt += topLine - (nLines - visLines);
	}

	if (scrollAmt == 0)
		return;

	/*
	 * Figure out if any of the currently visible lines can be reused
	 * by ScrollRect'ing them, and how many must be redrawn from scratch.
	 * firstRedraw is determined as a number from 0..visLines-1, then
	 * converted to an absolute line number.  This is because the value
	 * of topLine is volatile; firstRedraw isn't tied to it until it
	 * settles down.
	 */
	if (scrollAmt <= -visLines || scrollAmt >= visLines)
	{
		firstRedraw = 0;
		redrawCount = visLines;
	}
	else
	{
		if (scrollAmt > 0)
		{
			firstRedraw = 0;
			redrawCount = scrollAmt;
			scrollFrom = topLine;
		}
		else	/* scrollAmt < 0 */
		{
			firstRedraw = visLines + scrollAmt;
			redrawCount = -scrollAmt;
			scrollFrom = topLine - scrollAmt;
		}
		/*
		 * Don't scrollRect the stuff if not front window - might scroll in
		 * background pattern if stuff to be scrolled is under another window.
		 */
		if (theList->port == FrontWindow ())
			ScrollListLines (theList, scrollFrom, scrollFrom + scrollAmt);
		else
		{
			firstRedraw = 0;					/* draw from first visible */
			redrawCount = theList->visLines;	/* and draw all visible */
		}
	}
	theList->topVisLine -= scrollAmt;
	firstRedraw += theList->topVisLine;
	DrawListLines (theList, firstRedraw, firstRedraw + redrawCount - 1);
	ListSetScroll (theList);
}


/*
 * Scroll LineList object up or down n lines.  n < 0 means scroll up
 * (towards beginning).
 */

void
ListScrollRel (LListPtr theList, short n)
{
	if (n < 0)
		n += theList->topVisLine;
	else
		n += theList->topVisLine + theList->visLines - 1;
	ScrollToLine (theList, n);
}


/*
 * Set the scroll bar value
 */

void
ListSetScroll (LListPtr theList)
{
	SetCtlValue (theList->scroll, theList->topVisLine);
}


/*
 * Set the scroll bar maximum value and highlight appropriately.
 * It unhilited if there aren't enough lines to make the scroll bar
 * necessary, or if the list isn't currently activated.
 */

void
ListScrollMax (LListPtr theList)
{
ControlHandle	scroll;
short	maxScroll;

	scroll = theList->scroll;
	maxScroll = theList->nLines - theList->visLines;
	if (maxScroll <= 0)
	{
		maxScroll = 0;
	}
	SetCtlMax (scroll, maxScroll);
	HiliteControl (scroll, maxScroll > 0
							&& theList->hilite
							&& theList->port == FrontWindow ()
							? normalHilite : dimHilite);
}


/*
 * Select a line in the LineList.  The previous selection is
 * unhighlighted and the new selection is highlighted.
 * Selecting noLine unhilites currently selected line.
 *
 * If the list isn't currently activated (hilite = true), just
 * set the curLine field and do nothing.
 *
 * Note that InvertLine is smart about ignoring line "noLine" and
 * non-visible lines.
 */

void
SelectLine (LListPtr theList, short n)
{
short	prevCurLine;

	if (n < 0 || n >= theList->nLines)	/* trim to range */
		n = noLine;

	prevCurLine = theList->curLine;
	theList->curLine = n;

	if (theList->hilite && n != prevCurLine)
	{
		InvertLine (theList, prevCurLine);	/* turn off previous line */
		InvertLine (theList, n);			/* turn on new line */
	}
}


/*
 * ListTestScroll
 * Handle clicks in the scroll bar of a LineList object.  If it returns
 * false, the click was not in the scroll bar, otherwise it was and
 * scrolling has already been handled.
 *
 * ListTestText
 * Call this to see if the point was in the text rectangle of a
 * LineList object.  If not, ListTestText returns false.  Otherwise,
 * the line clicked on is highlighted and then the mouse is tracked
 * so that the line the mouse is in is hilited even if the mouse is
 * dragged.  If the mouse is dragged off the top or bottom of the
 * text rectangle, autoscrolling occurs.  If the user clicks or
 * drags below the bottom filled line but still in the text rectangle,
 * the current selection is deselected, but true is still returned.
 */


static pascal void
TrackScroll (ControlHandle theScroll, short partCode)
{
short	lDelta;

    if (partCode == trackPart)			/* still in same part? */
    {
        switch (partCode)
        {
            case inUpButton: lDelta = -1; break;
            case inDownButton: lDelta = 1; break;
            case inPageUp: lDelta = -(trackList->visLines-1); break;
            case inPageDown: lDelta = trackList->visLines-1; break;
        }
		ListScrollRel (trackList, lDelta);
    }
}


/*
 * Test whether the point lies in theList's scroll bar.  Handle scrolling
 * if so and return true, else return false.
 */

Boolean
ListTestScroll (LListPtr theList, Point pt)
{
ControlHandle	scrollBar;
Boolean			result = false;
short			line;

	scrollBar = theList->scroll;
	if ((trackPart = TestControl (scrollBar, pt)) != 0)
	{
		if (trackPart == inThumb)
		{
			line = GetCtlValue (scrollBar);
			if (TrackControl (scrollBar, pt, nil) == inThumb)
				ListScrollRel (theList, GetCtlValue (scrollBar) - line);
		}
		else
		{
			trackList = theList;
			(void) TrackControl (scrollBar, pt, TrackScroll);
		}
		result = true;
	}
	return (result);
}


/*
 * Test whether the point lies in theList's text rect.  If it does,
 * select the correct line.  Handles autoscrolling when the mouse is
 * dragged outside of the text rectangle.
 */

Boolean
ListTestText (LListPtr theList, Point pt)
{
Boolean	result = false;
short	line;
short	topLine;
short	bottomLine;

	if (PtInRect (pt, &theList->textRect))
	{
		for (;;)
		{
			line = FindLine (theList, pt);
			topLine = theList->topVisLine;
			bottomLine = topLine + theList->visLines - 1;
			if (line < topLine)
				line = topLine - 1;		/* scroll up one */
			if (line > bottomLine)
				line = bottomLine + 1;	/* scroll down one */
			if (line < 0)
				line = 0;				/* can only scroll to line zero */
			if (line > theList->nLines - 1)
			{
				if (PtInRect (pt, &theList->textRect))
					line = noLine;
				else
					line = theList->nLines - 1;
			}
			SelectLine (theList, line);
			ScrollToLine (theList, line);
			if (!StillDown ())
				break;
			GetMouse (&pt);
		}
		result = true;
	}
	return (result);
}


/*
 * Insert a line into a line list at position n.  The line is
 * selected but that doesn't cause it to be scrolled into view.
 * The list is redrawn appropriately, i.e. with hiliting if it's
 * turned on.
 *
 * Some intelligence is lacking to handle pathological things, so
 * that, for instance, inserting lines above the top visible line
 * will cause the stuff that's visible to be drawn shoved down a line.
 */

Boolean
InsertLine (LListPtr theList, LineHandle theLine, short n)
{
short		i;
LAHandle	lines;
Size		newSize;

	SelectLine (theList, noLine);	/* unhilite any previous selection */

	if (n >= theList->maxLines)		/* uh-oh - no more room! */
	{
		lines = theList->lines;
		/* try to make handle 1K bigger */
		newSize = GetHandleSize ((Handle) lines) + 1024L;
		SetHandleSize ((Handle) lines, newSize);
		if (GetHandleSize ((Handle) lines) != newSize)
			return (false);			/* can't make room */
		theList->maxLines = GetHandleSize ((Handle) lines) / sizeof (LineHandle);
	}

	for (i = theList->nLines; i > n; --i)	/* make empty slot */
		ListLine (theList, i) = ListLine (theList, i-1);
	ListLine (theList, n) = theLine;
	++theList->nLines;
	ListScrollMax (theList);
	DrawListLines (theList, n, n + theList->visLines - 1);
	/*DrawListLines (theList, n, theList->topVisLine + theList->visLines - 1);*/
	SelectLine (theList, n);
	return (true);
}


/*
 * Delete a line from a line list.
 */

void
DeleteLine (LListPtr theList, short n)
{
short	i;

	SelectLine (theList, noLine);		/* unhilite previous selection */
	DisposeLine (theList, n);			/* toss storage for line */
	for (i = n; i < theList->nLines - 1; ++i)
		ListLine (theList, i) = ListLine (theList, i+1);	/* close up gap */
	--theList->nLines;
	ListScrollMax (theList);
	DrawListText (theList);
	/*ScrollToLine (theList, n);*/
}


/*
 * Paste a new string into one of a line's field.  It's redrawn
 * as well.
 */

void
PasteField (LListPtr theList, short lineNo, short fieldNo, StringPtr str)
{
FieldHandle	hField;
short		i = fieldNo;

	hField = ListLine (theList, lineNo);
	while (i-- > 0)
		hField = NextField (hField);
	SetFieldStr (hField, str);
	DrawField (theList, lineNo, fieldNo);
}



/*
 * Paste a line into a line list.  It's redrawn as it's pasted,
 * and if it was the selected line before, is drawn hilited.
 *
 * This could all be much smarter (and faster) but I don't care.
 * It works, and the speedup wouldn't be enough to make it worth
 * it.
 */

void
PasteLine (LListPtr theList, LineHandle theLine, short n)
{
FieldHandle	hField;
short		i;

	for (i = 0, hField = theLine;
			hField != nil;
				hField = NextField (hField), ++i)
	{
		HLock ((Handle) hField);
		PasteField (theList, n, i, (**hField).fStr);
		HUnlock ((Handle) hField);
	}
}


/*
 * Turn list hiliting on or off.  This also affects the scroll bar.
 */

void
ListActivate (LListPtr theList, Boolean hilite)
{
	if (theList->hilite != hilite)		/* do this only on state change */
	{
		theList->hilite = hilite;
		InvertLine (theList, theList->curLine);
	}
	ListScrollMax (theList);
}


/*
 * Toss all the lines of a list and redraw.  This doesn't change
 * the list hiliting state.
 */

void
ResetList (LListPtr theList)
{
short	i;

	for (i = 0; i < theList->nLines; ++i)
		DisposeLine (theList, i);
	theList->nLines = 0;
	ListScrollMax (theList);
	ListSetScroll (theList);
	DrawListText (theList);
}


/* ---------------------------------------------------------------- */
/*					LineList drawing routines						*/
/* ---------------------------------------------------------------- */


/* ---------------------------------------------------------------- */
/*	There is some code redundancy among these routines.  The		*/
/*	redundancy is for the purpose of increasing drawing speed.		*/
/*	It is necessary to save and restore the port properly.  This	*/
/*	is only done in the routines where that is really necessary.	*/
/* ---------------------------------------------------------------- */



/* ---------------------------------------------------------------- */
/*					Display area calculations						*/
/* ---------------------------------------------------------------- */


/*
 * Get the rectangle enclosing the display area for one of the
 * lines.  The line number is an index in the range 0..visLines-1.
 * The left and right boundaries are the same as the general text
 * display rectangle, so only need to adjust top and bottom.
 */

void
GetLineRect (Rect *r, LListPtr theList, short lineNo)
{
short	height;

	*r = theList->textRect;
	height = theList->lineHeight;
	r->top += lineNo * height;
	r->bottom = r->top + height;
}


/*
 * Get the rectangle enclosing the display area for one of the
 * fields of a line.  The line number is an index in the range
 * 0..visLines-1.  The field number is an index in the range
 * 0..nFields-1.
 *
 * Must offset horizontally to get the absolute window position
 * since the offsets are relative to the text rectangle.
 * Leave one space to account for the separator line to the
 * left of the next field.
 */

void
GetFieldRect (Rect *r, LListPtr theList, short lineNo, short fieldNo)
{
	GetLineRect (r, theList, lineNo);
	r->left = theList->offset[fieldNo];
	r->right = theList->offset[fieldNo + 1] - 1;
	OffsetRect (r, theList->textRect.left, 0);
}


/* ---------------------------------------------------------------- */
/*							Drawing routines						*/
/* ---------------------------------------------------------------- */


/*
 * Scroll the list's text rectangle.  Note that this may not produce
 * the correct effect if the window is not the front window - some
 * of the stuff shifted might not be visible, and when shifted out from
 * under another window, shifts out as the background pattern.
 */

void
ScrollListLines (LListPtr theList, short from, short to)
{
GrafPtr		curPort;
RgnHandle	rgn;

	GetPort (&curPort);
	SetPort (theList->port);
	rgn = NewRgn ();
	ScrollRect (&theList->textRect, 0, (to-from) * theList->lineHeight, rgn);
	DisposeRgn (rgn);
	SetPort (curPort);
}


/*
 * Check whether a line is currently visible in the list's text
 * rectangle.
 */

Boolean
LineVisible (LListPtr theList, short lineNo)
{
	return (lineNo != noLine && lineNo >= theList->topVisLine
			&& lineNo < theList->topVisLine + theList->visLines);
}


/*
 * Invert a line from a list.  The line number passed is not
 * assumed to be visible.
 *
 * This routine will give you trouble if you're not careful!
 */

void
InvertLine (LListPtr theList, short lineNo)
{
GrafPtr	curPort;
short	i;
Rect	r;

	if (!LineVisible (theList, lineNo))
			return;
	GetPort (&curPort);
	SetPort (theList->port);
	GetLineRect (&r, theList, lineNo - theList->topVisLine);
	InvertRect (&r);
	SetPort (curPort);
}


/*
 * Draw a field from a list.  The line number passed is not assumed to
 * be visible.  If the line doesn't really exist (i.e., lineNo >= nLines)
 * a blank field is drawn.
 */

void
DrawField (LListPtr theList, short lineNo, short fieldNo)
{
GrafPtr		curPort;
Rect		r;
FieldHandle	hField;
Boolean		invert;
RgnHandle	rgn;
short		textMode;

	if (!LineVisible (theList, lineNo))
			return;

	GetPort (&curPort);
	SetPort (theList->port);
	invert = theList->hilite && (theList->curLine == lineNo);
	GetFieldRect (&r, theList, lineNo - theList->topVisLine, fieldNo);
	FillRect (&r, invert ? black : white);
	if (lineNo < theList->nLines)
	{
		hField = ListLine (theList, lineNo);
		while (fieldNo-- > 0)
			hField = NextField (hField);
		rgn = NewRgn ();
		GetClip (rgn);
		ClipRect (&r);
		MoveTo (r.left + 2, r.bottom - 4);
		textMode = theList->port->txMode;
		TextMode (invert? srcBic : srcOr);
		HLock ((Handle) hField);
		DrawString ((**hField).fStr);
		HUnlock ((Handle) hField);
		TextMode (textMode);
		SetClip (rgn);
		DisposeRgn (rgn);
	}
	SetPort (curPort);
}


/*
 * Draw a line from a list.  The line number passed is not assumed to be
 * visible.  If the line doesn't really exist (i.e., lineNo >= nLines)
 * then draw blank fields.
 */

void
DrawLine (LListPtr theList, short lineNo)
{
GrafPtr	curPort;
short	i;
short	topLine;
short	top, bottom, left;
short	penMode;
Rect	r;
FieldHandle	hField;
Boolean		invert;

	if (!LineVisible (theList, lineNo))
			return;

	GetPort (&curPort);
	SetPort (theList->port);
	topLine = theList->topVisLine;
	
	invert = theList->hilite && (lineNo == theList->curLine);

	top = theList->textRect.top + theList->lineHeight * (lineNo - topLine);
	bottom = top + theList->lineHeight - 1;
	penMode = theList->port->pnMode;
	if (invert)
		PenMode (patBic);

	for (i = 0; i < theList->nFields; ++i)
	{
		DrawField (theList, lineNo, i);
		if (i > 0)
		{
			left = theList->textRect.left + theList->offset[i] - 1;
			MoveTo (left, top);
			LineTo (left, bottom);
		}
	}

	PenMode (penMode);
	SetPort (curPort);
}


void
DrawListLines (LListPtr theList, short from, short to)
{
short	i;

	for (i = from;  i <= to; ++i)
	{
		DrawLine (theList, i);
	}
}


void
DrawListText (LListPtr theList)
{
	DrawListLines (theList, theList->topVisLine,
					theList->topVisLine + theList->visLines - 1);
}


void
DrawListFrame (LListPtr theList)
{
GrafPtr	curPort;
Rect	r;
short	i;
short	offset;

	GetPort (&curPort);
	SetPort (theList->port);
	r = theList->textRect;
	InsetRect (&r, -1, -1);
	FrameRect (&r);
	InsetRect (&r, 1, 1);

	/*
	 * Draw separator lines.  These are drawn one pixel to the left of
	 * the field area.
	 */

	for (i = 1; i < theList->nFields; ++i)
	{
		offset = r.left + theList->offset[i] - 1;
		MoveTo (offset, r.top);
		LineTo (offset, r.bottom);
	}
	SetPort (curPort);
}


/*
 * Redraw the list.  Set the control hiliting value to its current
 * value to redraw it.  Draw the text area frame (including field
 * separating bars), then draw all the visible lines.
 */

void
DrawList (LListPtr theList)
{
ControlHandle	scroll;

	scroll = theList->scroll;
	HiliteControl (scroll, (**scroll).contrlHilite);
	DrawListFrame (theList);
	DrawListText (theList);
}
